#!/usr/bin/perl -w
# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Kagetaka Libraries.
#
# The Initial Developer of the Original Code is Hizuya Atsuzaki
# Portions created by the Initial Developer are Copyright (C) 2003
# the Initial Developer. All Rights Reserved.
#
# Contributor(s): Hizuya Atsuzaki <hizuya@hizlab.net>
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****

use Socket;
use English;
use POSIX qw(sys_wait_h);
use IO::Handle;
use Digest::MD5 qw(md5_base64);
use strict;

my $server_addr = "127.0.0.1";
my $server_port = 8080;

use constant MON  => qw(Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec);
use constant WDAY => qw(Sun Mon Tue Wed Thu Fri Sat);

use constant HTTP_DEFAULT_PORT => 80;

### reaper for child process
sub reaper_child() {
	waitpid(-1, &WNOHANG);
	$SIG{CHLD} = \&reaper_child;
}

### interrupted
sub interrupted() {
	close(CLIENT_WAITING);
	exit(0);
}

### send to client
sub send($$) {
	my ($client, $line) = @_;
	print STDOUT " $client->{key} : Response : [$line]\r\n";
	print CLIENT "$line\r\n";
}

### parse header
sub parse_header(@) {
	my(@headers) = @_;
	my %table;
	foreach my $header (@headers) {
		my($name, $value) = split(/:[ ]*/, $header, 2);
		if (!defined($value)) {  # request line...
			$value = $name;
			$name  = "";
		}
		$table{$name} = $value;
	}
	return %table;
}

### parse query
sub parse_query($) {
	my($client) = @_;
	my @pairs = split(/&/, $client->{query});
	my %table;
	foreach my $pair (@pairs) {
		$pair =~ s/\+/ /g;
		$pair =~ s/%(..)/pack("c",hex($1))/ge;
		my($name, $value) = split(/=/, $pair, 2);
		$table{$name} = $value;
	}
	return %table;
}

### send file
sub send_file($@) {
	my($client, @headers) = @_;
	my($path) = ($client->{path} =~ m@^/(.*)$@);

	# index file
	if (length($path) == 0 || $path =~ m@/$@) {
		$path .= "index.html";
	}

	if (-f $path) {
		my $content_type = "text/palin";
		if ($path =~ /\.htm[l]?$/) {
			$content_type = "text/html";
		} elsif ($path =~ /\.gif$/) {
			$content_type = "image/gif";
		} elsif ($path =~ /\.(?:jpeg|jpg)$/) {
			$content_type = "image/jpeg";
		} elsif ($path =~ /\.png$/) {
			$content_type = "image/png";
		}

		my %headers = &parse_header(@headers);
		my $ifModifiedSince = $headers{'If-Modified-Since'};
		my $ifNoneMatch     = $headers{'If-None-Match'    };

		my @stat = stat($path);
		my $filelen = $stat[7];
		my($s_sec, $s_min, $s_hour, $s_mday, $s_mon, $s_year, $s_wday) = gmtime($stat[9]);
		my $lastModified = sprintf("%s, %02d %s %04d %02d:%02d:%02d GMT",
		                           (WDAY)[$s_wday], $s_mday, (MON)[$s_mon], $s_year + 1900, $s_hour, $s_min, $s_sec);
		my $eTag = md5_base64($path.$lastModified); # umuu...

		my $code = "200 Ok";
		my $send = 1;

		if (defined($ifModifiedSince) || defined($ifNoneMatch)) {
			# umuu....
			if ((!defined($ifModifiedSince) || ($ifModifiedSince eq $lastModified))
			 && (!defined($ifNoneMatch    ) || ($ifNoneMatch     eq $eTag        ))) {
				$code = "304 Not Modified";
				$send = 0;
			}
		}

		&send($client, "HTTP/1.0 $code");
		&send($client, "Connection: close");
		&send($client, "Content-Type: $content_type");
		&send($client, "Content-Length: $filelen");
		&send($client, "Last-Modified: $lastModified");
		&send($client, "ETag: $eTag");
		&send($client, "");

		if ($send) {
			open(READFILE, "<$path")
				|| die("Can not open file($path). $ERRNO");

			my $data;
			my $rlen;
			while (defined($rlen = sysread(READFILE, $data, $filelen))) {
				$filelen -= $rlen;
				syswrite(CLIENT, $data, $rlen);
				last if ($filelen <= 0);
			}
			close(READFILE);
		}

		return;
	} elsif (-d $path) {
		my %headers = &parse_header(@headers);
		my $host    = $headers{'Host'};
		if (!defined($host)) {
			$host = "$server_addr";
			if ($server_port != HTTP_DEFAULT_PORT) {
				$host .= ":$server_port";
			}
		}

		&send($client, "HTTP/1.0 302 Moved Temporarily");
		&send($client, "Location: http://$host/$path/");
		&send($client, "Connection: close");
		&send($client, "Content-Length: 0");
		&send($client, "");

		return;
	}

	&send($client, "HTTP/1.0 404 Not found");
	&send($client, "Connection: close");
	&send($client, "Content-Type: text/html");
	&send($client, "");
	&send($client, "<html>");
	&send($client, "<head><title>404 Not found</title></head>");
	&send($client, "<body>");
	&send($client, "<h1>404 Not found</h1>");
	&send($client, "<p><code>[$client->{path}]</code> not found.</p>");
	&send($client, "<body></html>");
}

### dump.cgi
sub cgi_dump($$@) {
	my($client, $length, @headers) = @_;
	my $count = $#headers;
	my @lines = @headers;

	# POST method
	if ($client->{method} =~ m/POST/i) {
		my $data;
		my $rlen;
		$lines[$count++] = "--- POST START ---";
		$lines[$count] = "";
		while (defined($rlen = read(CLIENT, $data, $length))) {
			$length -= $rlen;
			$lines[$count] = "$lines[$count]$data";
			last if ($length <= 0);
		}
		$count++;
		$lines[$count++] = "--- POST END ---";
	}

	&send($client, "HTTP/1.0 200 Ok");
	&send($client, "Connection: close");
	&send($client, "Content-Type: text/plain");
	&send($client, "");
	for (my $i = 0; $i < $count; $i++) {
		&send($client, "[$lines[$i]]");
	}
}

### wait.cgi
sub cgi_wait($$@) {
	my($client, $length, @headers) = @_;
	my %query = &parse_query($client);
	my $time  = $query{'time'};
	my $path  = $query{'path'};

	if (defined($time) && $time =~ /^[0-9]+$/) {
		sleep($time);
	}

	if (!defined($path)) {
		$path = "";
	}
	$client->{path} = $path;
	&send_file($client, @headers);
}

### main
sub main() {
	my $bind = inet_aton($server_addr);

	socket(CLIENT_WAITING, PF_INET, SOCK_STREAM, 0)
		|| die "Can not create socket. $ERRNO";

	setsockopt(CLIENT_WAITING, SOL_SOCKET, SO_REUSEADDR, 1)
		|| die "Can not set socket options. $ERRNO";

	bind(CLIENT_WAITING, pack_sockaddr_in($server_port, $bind))
		|| die "Can not bind to socket. $ERRNO";

	listen(CLIENT_WAITING, SOMAXCONN)
		|| die "Can not listen from socket. $ERRNO";

	if ($OSNAME ne "MSWin32") {
		$SIG{CHLD} = \&reaper_child;
		$SIG{INT}  = \&interrupted;
	}
	STDOUT->autoflush(1);

	print "Accept on $server_addr:$server_port\r\n";
	while (1) {
		my $paddr = accept(CLIENT, CLIENT_WAITING);

		my $child_pid = fork();
		if (!defined($child_pid)) {
			die "Can not fork.";
		}

		if ($child_pid != 0) {
			# Parent process
			close(CLIENT);
			next;
		}

		# Chile process

		my $client = {
			key      => undef,
			iaddr    => undef,
			hostname => undef,
			ip       => undef,
			port     => undef,
			method   => undef,
			url      => undef,
			version  => undef,
			path     => undef,
			query    => undef,
		};

		($client->{port}, $client->{iaddr}) = unpack_sockaddr_in($paddr);
		$client->{hostname} = gethostbyaddr($client->{iaddr}, AF_INET);
		$client->{ip}       = inet_ntoa($client->{iaddr});
		$client->{key}      = "$client->{ip}:$client->{port}";

		print " $client->{key} : Connect\r\n";

		CLIENT->autoflush(1);

		my $count   = 0;
		my @headers = ();
		my $length  = 0;

		while (<CLIENT>) {
			## Header
			s/\r\n$//;
			if ($count == 0 && "$_" eq "") {
				next;
			}

			$headers[$count++] = $_;

			if ($count == 1) {
				($client->{method}, $client->{url}, $client->{version}) = split(/ +/, $_, 3);
			} elsif (/^Content-Length:[ ]*(.*)/i) {
				$length = $1;
			} elsif ("$_" eq "") {
				last;
			}

			print " $client->{key} : Header : [$_]\r\n";
		}

		## Body
		($client->{path}, $client->{query}) = ($client->{url} =~ m@(^[^\?]*)[\?]?(.*)$@);

		RESPONSE:
		{
			# CGI
			if ($client->{path} =~ m@/([^/\.]*)\.cgi$@) {
				if (defined(eval("&cgi_$1(".'$client, $length, @headers);'))) {
					last RESPONSE;
				}
			}

			# others
			&send_file($client, @headers);
		}

		print " $client->{key} : Close\r\n";
		close(CLIENT);
		return;
	}
}

&main;
